@@ -1,58 +1,64 @@
 package Catalyst::Authentication::Credential::OpenID;
+use warnings;
 use strict;
-# use warnings; no warnings "uninitialized"; # for testing, not production
-use parent "Class::Accessor::Fast";
+use base "Class::Accessor::Fast";
 
-BEGIN {
-    __PACKAGE__->mk_accessors(qw/ _config realm debug secret /);
-}
+__PACKAGE__->mk_accessors(qw/
+    realm debug secret
+    openid_field
+    consumer_secret
+    ua_class
+    ua_args
+    extension_args
+    errors_are_fatal
+    extensions
+    trust_root
+/);
 
-our $VERSION = "0.14_05";
+our $VERSION = "0.16_01";
 
 use Net::OpenID::Consumer;
 use Catalyst::Exception ();
 
-sub new : method {
+sub new {
     my ( $class, $config, $c, $realm ) = @_;
-    my $self = { _config => { %{ $config },
-                              %{ $realm->{config} }
-                          }
-                 };
+    my $self = {
+                %{ $config },
+                %{ $realm->{config} }
+               };
     bless $self, $class;
 
     # 2.0 spec says "SHOULD" be named "openid_identifier."
-    $self->_config->{openid_field} ||= "openid_identifier";
+    $self->{openid_field} ||= "openid_identifier";
 
-    $self->debug( $self->_config->{debug} );
-
-    my $secret = $self->_config->{consumer_secret} ||= join("+",
-                                                            __PACKAGE__,
-                                                            $VERSION,
-                                                            sort keys %{ $c->config }
-                                                            );
+    my $secret = $self->{consumer_secret} ||= join("+",
+                                                   __PACKAGE__,
+                                                   $VERSION,
+                                                   sort keys %{ $c->config }
+                                                  );
 
     $secret = substr($secret,0,255) if length $secret > 255;
     $self->secret($secret);
     # If user has no preference we prefer L::PA b/c it can prevent DoS attacks.
-    $self->_config->{ua_class} ||= eval "use LWPx::ParanoidAgent" ?
+    my $ua_class = $self->{ua_class} ||= eval "use LWPx::ParanoidAgent" ?
         "LWPx::ParanoidAgent" : "LWP::UserAgent";
 
-    my $agent_class = $self->_config->{ua_class};
+    my $agent_class = $self->ua_class;
     eval "require $agent_class"
         or Catalyst::Exception->throw("Could not 'require' user agent class " .
-                                      $self->_config->{ua_class});
+                                      $self->ua_class);
 
     $c->log->debug("Setting consumer secret: " . $secret) if $self->debug;
 
     return $self;
 }
 
-sub authenticate : method {
+sub authenticate {
     my ( $self, $c, $realm, $authinfo ) = @_;
 
     $c->log->debug("authenticate() called from " . $c->request->uri) if $self->debug;
 
-    my $field = $self->{_config}->{openid_field};
+    my $field = $self->openid_field;
 
     my $claimed_uri = $authinfo->{ $field };
 
@@ -62,44 +68,46 @@ sub authenticate : method {
 
 
     my $csr = Net::OpenID::Consumer->new(
-        ua => $self->_config->{ua_class}->new(%{$self->_config->{ua_args} || {}}),
+        ua => $self->ua_class->new(%{$self->ua_args || {}}),
         args => $c->req->params,
         consumer_secret => $self->secret,
     );
 
-    if ( $self->_config->{extension_args} and $self->debug )
+    if ( $self->extension_args )
     {
-        $c->log->info("The configuration key 'extension_args' is deprecated; use 'extensions'");
+        $c->log->warn("The configuration key 'extension_args' is ignored; use 'extensions'");
     }
 
-    my @extensions = $self->_config->{extensions} ?
-        @{ $self->_config->{extensions} } : $self->_config->{extension_args} ?
-        @{ $self->_config->{extension_args} } : ();
+    my %extensions = ref($self->extensions) eq "HASH" ?
+        %{ $self->extensions } : ref($self->extensions) eq "ARRAY" ?
+        @{ $self->extensions } : ();
 
     if ( $claimed_uri )
     {
-        my $current = $c->uri_for($c->req->uri->path); # clear query/fragment...
+        my $current = $c->uri_for("/" . $c->req->path); # clear query/fragment...
 
         my $identity = $csr->claimed_identity($claimed_uri);
         unless ( $identity )
         {
-            if ( $self->_config->{errors_are_fatal} )
+            if ( $self->errors_are_fatal )
             {
                 Catalyst::Exception->throw($csr->err);
             }
             else
             {
                 $c->log->error($csr->err . " -- $claimed_uri");
-                $c->detach();
+                return;
             }
         }
 
-        $identity->set_extension_args(@extensions)
-            if @extensions;
+        for my $key ( keys %extensions )
+        {
+            $identity->set_extension_args($key, $extensions{$key});
+        }
 
         my $check_url = $identity->check_url(
             return_to  => $current . '?openid-check=1',
-            trust_root => $current,
+            trust_root => $self->trust_root || $current,
             delayed_return => 1,
         );
         $c->res->redirect($check_url);
@@ -122,8 +130,8 @@ sub authenticate : method {
             my $user = +{ map { $_ => scalar $identity->$_ }
                 qw( url display rss atom foaf declared_rss declared_atom declared_foaf foafmaker ) };
             # Dude, I did not design the array as hash spec. Don't curse me [apv].
-            my %flat = @extensions;
-            for my $key ( keys %flat )
+
+            for my $key ( keys %extensions )
             {
                 $user->{extensions}->{$key} = $identity->signed_extension_fields($key);
             }
@@ -142,7 +150,7 @@ sub authenticate : method {
         }
         else
         {
-            $self->_config->{errors_are_fatal} ?
+            $self->errors_are_fatal ?
                 Catalyst::Exception->throw("Error validating identity: " . $csr->err)
                       :
                 $c->log->error( $csr->err);
@@ -161,13 +169,13 @@ Catalyst::Authentication::Credential::OpenID - OpenID credential for Catalyst::P
 
 =head1 VERSION
 
-0.14_05
+0.16_01
 
 =head1 BACKWARDS COMPATIBILITY CHANGES
 
-=head2 EXTENTION_ARGS v EXTENSIONS
+=head2 EXTENSION_ARGS v EXTENSIONS
 
-B<NB>: The extenstions were previously configured under the key C<extension_args>. They are now configured under C<extensions>. This prevents the need for double configuration but it breaks extensions in your application if you do not change the name. The old version is supported for now but may be phased out at any time.
+B<NB>: The extensions were previously configured under the key C<extension_args>. They are now configured under C<extensions>. C<extension_args> is no longer honored.
 
 As previously noted, L</EXTENSIONS TO OPENID>, I have not tested the extensions. I would be grateful for any feedback or, better, tests.
 
@@ -194,8 +202,8 @@ Somewhere in myapp.conf-
          <openid>
              <credential>
                  class   OpenID
+                 ua_class   LWP::UserAgent
              </credential>
-             ua_class   LWP::UserAgent
          </openid>
      </realms>
  </Plugin::Authentication>
@@ -208,7 +216,7 @@ Or in your myapp.yml if you're using L<YAML> instead-
      openid:
        credential:
          class: OpenID
-       ua_class: LWP::UserAgent
+         ua_class: LWP::UserAgent
 
 In a controller, perhaps C<Root::openid>-
 
@@ -353,24 +361,25 @@ clear text passwords and one called "openid" which uses... uh, OpenID.
                           }
               },
               openid => {
-                  consumer_secret => "Don't bother setting",
-                  ua_class => "LWP::UserAgent",
-                  ua_args => {
-                      whitelisted_hosts => [qw/ 127.0.0.1 localhost /],
-                  },
                   credential => {
                       class => "OpenID",
                       store => {
                           class => "OpenID",
                       },
-                  },
-                  extensions => [
-                      'http://openid.net/extensions/sreg/1.1',
-                      {
-                       required => 'email',
-                       optional => 'fullname,nickname,timezone',
+                      consumer_secret => "Don't bother setting",
+                      ua_class => "LWP::UserAgent",
+                      # whitelist is only relevant for LWPx::ParanoidAgent
+                      ua_args => {
+                          whitelisted_hosts => [qw/ 127.0.0.1 localhost /],
                       },
-                  ],
+                      extensions => [
+                          'http://openid.net/extensions/sreg/1.1',
+                          {
+                           required => 'email',
+                           optional => 'fullname,nickname,timezone',
+                          },
+                      ],
+                  },
               },
           },
       }
@@ -398,23 +407,23 @@ This is the same configuration in the default L<Catalyst> configuration format f
              </credential>
          </members>
          <openid>
-             <ua_args>
-                 whitelisted_hosts   127.0.0.1
-                 whitelisted_hosts   localhost
-             </ua_args>
-             consumer_secret   Don't bother setting
-             ua_class   LWP::UserAgent
              <credential>
                  <store>
                      class   OpenID
                  </store>
                  class   OpenID
+                 <ua_args>
+                     whitelisted_hosts   127.0.0.1
+                     whitelisted_hosts   localhost
+                 </ua_args>
+                 consumer_secret   Don't bother setting
+                 ua_class   LWP::UserAgent
+                 <extensions>
+                     http://openid.net/extensions/sreg/1.1
+                     required   email
+                     optional   fullname,nickname,timezone
+                 </extensions>
              </credential>
-             <extensions>
-                 http://openid.net/extensions/sreg/1.1
-                 required   email
-                 optional   fullname,nickname,timezone
-             </extensions>
          </openid>
      </realms>
  </Plugin::Authentication>
@@ -440,42 +449,73 @@ And now, the same configuration in L<YAML>. B<NB>: L<YAML> is whitespace sensiti
          class: OpenID
          store:
            class: OpenID
-       consumer_secret: Don't bother setting
-       ua_class: LWP::UserAgent
-       ua_args:
-         whitelisted_hosts:
-           - 127.0.0.1
-           - localhost
-       extensions:
-           - http://openid.net/extensions/sreg/1.1
-           - required: email
-             optional: fullname,nickname,timezone
+         consumer_secret: Don't bother setting
+         ua_class: LWP::UserAgent
+         ua_args:
+           # whitelist is only relevant for LWPx::ParanoidAgent
+           whitelisted_hosts:
+             - 127.0.0.1
+             - localhost
+         extensions:
+             - http://openid.net/extensions/sreg/1.1
+             - required: email
+               optional: fullname,nickname,timezone
 
 B<NB>: There is no OpenID store yet.
 
+You can set C<trust_root> now too. This is experimental and I have no idea if it's right or could be better. Right now it must be a URI. It was submitted as a path but this seems to limit it to the Catalyst app and while easier to dynamically generate no matter where the app starts, it seems like the wrong way to go. Let me know if that's mistaken.
+
 =head2 EXTENSIONS TO OPENID
 
 The Simple Registration--L<http://openid.net/extensions/sreg/1.1>--(SREG) extension to OpenID is supported in the L<Net::OpenID> family now. Experimental support for it is included here as of v0.12. SREG is the only supported extension in OpenID 1.1. It's experimental in the sense it's a new interface and barely tested. Support for OpenID extensions is here to stay.
 
-=head2 MORE ON CONFIGURATION
+Google's OpenID is also now supported. Uh, I think.
 
-These are set in your realm. See above.
+Here is a snippet from Thorben JE<auml>ndling combining Sreg and Google's extenstionsE<ndash>
+
+ 'Plugin::Authentication' => {
+    openid => {
+        credential => {
+            class => 'OpenID',
+            ua_class => 'LWP::UserAgent',
+            extensions => {
+                'http://openid.net/extensions/sreg/1.1' => {
+                    required => 'nickname,email,fullname',
+                    optional => 'timezone,language,dob,country,gender'
+                },
+                'http://openid.net/srv/ax/1.0' => {
+                    mode => 'fetch_request',
+                    'type.nickname' => 'http://axschema.org/namePerson/friendly',
+                    'type.email' => 'http://axschema.org/contact/email',
+                    'type.fullname' => 'http://axschema.org/namePerson',
+                    'type.firstname' => 'http://axschema.org/namePerson/first',
+                    'type.lastname' => 'http://axschema.org/namePerson/last',
+                    'type.dob' => 'http://axschema.org/birthDate',
+                    'type.gender' => 'http://axschema.org/person/gender',
+                    'type.country' => 'http://axschema.org/contact/country/home',
+                    'type.language' => 'http://axschema.org/pref/language',
+                    'type.timezone' => 'http://axschema.org/pref/timezone',
+                    required => 'nickname,fullname,email,firstname,lastname',
+                    if_available => 'dob,gender,country,language,timezone',
+            },
+            },
+        },
+    },
+    default_realm => 'openid',
+ };
+
+
+=head2 MORE ON CONFIGURATION
 
 =over 4
 
 =item ua_args and ua_class
 
-L<LWPx::ParanoidAgent> is the default agent E<mdash> C<ua_class> E<mdash> if it's available, L<LWP::UserAgent> if not. You don't have to set
-it. I recommend that you do B<not> override it. You can with any well behaved L<LWP::UserAgent>. You probably should not.
-L<LWPx::ParanoidAgent> buys you many defenses and extra security checks. When you allow your application users freedom to initiate external
-requests, you open an avenue for DoS (denial of service) attacks. L<LWPx::ParanoidAgent> defends against this. L<LWP::UserAgent> and any
-regular subclass of it will not.
+L<LWPx::ParanoidAgent> is the default agent E<mdash> C<ua_class> E<mdash> if it's available, L<LWP::UserAgent> if not. You don't have to set it. I recommend that you do B<not> override it. You can with any well behaved L<LWP::UserAgent>. You probably should not. L<LWPx::ParanoidAgent> buys you many defenses and extra security checks. When you allow your application users freedom to initiate external requests, you open an avenue for DoS (denial of service) attacks. L<LWPx::ParanoidAgent> defends against this. L<LWP::UserAgent> and any regular subclass of it will not.
 
 =item consumer_secret
 
-The underlying L<Net::OpenID::Consumer> object is seeded with a secret. If it's important to you to set your own, you can. The default uses
-this package name + its version + the sorted configuration keys of your Catalyst application (chopped at 255 characters if it's longer).
-This should generally be superior to any fixed string.
+The underlying L<Net::OpenID::Consumer> object is seeded with a secret. If it's important to you to set your own, you can. The default uses this package name + its version + the sorted configuration keys of your Catalyst application (chopped at 255 characters if it's longer). This should generally be superior to any fixed string.
 
 =back
 
@@ -485,8 +525,7 @@ Option to suppress fatals.
 
 Support more of the new methods in the L<Net::OpenID> kit.
 
-There are some interesting implications with this sort of setup. Does a user aggregate realms or can a user be signed in under more than one
-realm? The documents could contain a recipe of the self-answering OpenID end-point that is in the tests.
+There are some interesting implications with this sort of setup. Does a user aggregate realms or can a user be signed in under more than one realm? The documents could contain a recipe of the self-answering OpenID end-point that is in the tests.
 
 Debug statements need to be both expanded and limited via realm configuration.
 
